from time import sleep
from typing import Any
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver import ActionChains
from selenium.webdriver.support.ui import Select
from selenium.webdriver.remote.webelement import WebElement

from fastrpa.core.file import File
from fastrpa.exceptions import FileNotDownloadableException, FormException
from fastrpa.utils import (
    find_element,
    find_elements,
    print_list,
    print_table,
)
from fastrpa.types import WebDriver


class Element:
    def __init__(self, xpath: str, webdriver: WebDriver) -> None:
        self.xpath = xpath
        self.webdriver = webdriver
        self.file = File(self.webdriver)
        self.actions = ActionChains(self.webdriver)

    @property
    def source(self) -> WebElement:
        return find_element(self.webdriver, self.xpath)

    def attribute(self, name: str) -> str | None:
        return self.source.get_attribute(name)

    @property
    def tag(self) -> str:
        return self.source.tag_name

    @property
    def id(self) -> str | None:
        return self.attribute('id')

    @property
    def css_class(self) -> list[str] | None:
        if classes := self.attribute('class'):
            return classes.split(' ')
        return []

    @property
    def css_inline(self) -> dict[str, str]:
        to_return = {}
        if style := self.attribute('style'):
            for css in style.strip().split(';'):
                if css:
                    css_property, css_value = css.split(':')
                    to_return[css_property.strip()] = css_value.strip()
        return to_return

    @property
    def text(self) -> str | None:
        if self.source.text:
            return self.source.text
        elif value := self.attribute('value'):
            return value
        return None

    @property
    def is_visible(self) -> bool:
        return self.source.is_displayed()

    def focus(self):
        self.actions.scroll_to_element(self.source)
        self.actions.move_to_element(self.source)
        self.actions.perform()

    def click(self):
        self.focus()
        self.actions.click(self.source)
        self.actions.perform()

    def check(self, attribute: str, value: str) -> bool:
        return self.attribute(attribute) == value


class InputElement(Element):
    def clear(self):
        self.source.clear()

    def fill(self, value: str):
        self.source.send_keys(value)

    def fill_slowly(self, value: str, delay: float = 0.3):
        for key in value:
            self.source.send_keys(key)
            sleep(delay)


class FileInputElement(Element):
    def attach_file(self, path: str):
        self.source.send_keys(self.file.get_path(path))


class SelectElement(Element):
    @property
    def select_source(self) -> Select:
        return Select(self.source)

    @property
    def options_values(self) -> list[str | None]:
        return [
            option.get_attribute('value')
            for option in self.select_source.options
        ]

    @property
    def options_labels(self) -> list[str | None]:
        return [
            option.get_attribute('innerText')
            for option in self.select_source.options
        ]

    @property
    def options(self) -> dict[str | None, str | None]:
        return {
            option.get_attribute('value'): option.get_attribute('innerText')
            for option in self.select_source.options
        }

    @property
    def current(self) -> tuple[str | None, str | None]:
        if value_id := self.attribute('value'):
            return value_id, self.options[value_id]
        return None, None

    def select(self, label: str | None = None, value: str | None = None):
        if label:
            self.select_source.select_by_visible_text(label)
        elif value:
            self.select_source.select_by_value(value)
        else:
            raise ValueError('You must provide at least "label" or "value"!')

    def has_option(self, label: str | None = None, value: str | None = None):
        if label:
            return label in self.options_labels
        elif value:
            return value in self.options_values
        raise ValueError('You must provide at least "label" or "value"!')

    def __contains__(self, key: Any) -> bool:
        return self.has_option(label=key) or self.has_option(value=key)

    def print(self):
        identifier = f'[@id="{self.id}"]' if self.id else self.xpath
        print_list(f'{identifier}', self.options)


class RadioInputElement(Element):
    @property
    def radio_sources(self) -> list[WebElement]:
        name = self.attribute('name')
        xpath = f'//input[@type="radio"][@name="{name}"][not(@disabled)]'
        return find_elements(self.webdriver, xpath)

    @property
    def options_values(self) -> list[str]:
        return [
            value
            for element in self.radio_sources
            if (value := element.get_attribute('value'))
        ]

    @property
    def options_labels(self) -> list[str]:
        return [
            find_element(self.webdriver, xpath).text
            for element in self.radio_sources
            if (xpath := f'//label[@for="{element.get_attribute("id")}"]')
        ]

    @property
    def options(self) -> dict[str, str]:
        return {
            value: label
            for value, label in zip(self.options_values, self.options_labels)
        }

    def select(self, label: str | None = None, value: str | None = None):
        if label:
            index = self.options_labels.index(label)
            element = self.radio_sources[index]
            self.actions.move_to_element(element)
            self.actions.click(element)
            self.actions.perform()

        elif value:
            index = self.options_values.index(value)
            element = self.radio_sources[index]
            self.actions.move_to_element(element)
            self.actions.click(element)
            self.actions.perform()
        else:
            raise ValueError('You must provide at least "label" or "value"!')

    def has_option(self, label: str | None = None, value: str | None = None):
        if label:
            return label in self.options_labels
        elif value:
            return value in self.options_values
        raise ValueError('You must provide at least "label" or "value"!')

    def __contains__(self, key: Any) -> bool:
        return self.has_option(label=key) or self.has_option(value=key)

    def print(self):
        identifier = f'[@id="{self.id}"]' if self.id else self.xpath
        print_list(f'{identifier}', self.options)


class CheckboxElement(Element):
    @property
    def is_checked(self) -> bool:
        return self.source.is_selected()

    def check(self):
        if not self.is_checked:
            self.click()

    def uncheck(self):
        if self.is_checked:
            self.click()

    def switch(self):
        self.click()


class ListElement(Element):
    @property
    def items_sources(self) -> list[WebElement]:
        return find_elements(self.webdriver, './/li')

    @property
    def is_ordered(self) -> bool:
        return self.tag == 'ol'

    @property
    def items(self) -> dict[str | None, str | None]:
        return {
            item.get_attribute('id'): item.get_attribute('innerText')
            for item in self.items_sources
        }

    @property
    def items_ids(self) -> list[str | None]:
        return [item.get_attribute('ids') for item in self.items_sources]

    @property
    def items_labels(self) -> list[str | None]:
        return [item.get_attribute('innerText') for item in self.items_sources]

    def click_in_item(self, label: str | None = None, id: str | None = None):
        if not (label or id):
            raise ValueError('You must provide at least "label" or "id"!')

        for item in self.items_sources:
            if label and label == item.get_attribute('innerText'):
                self.actions.click(item)
                self.actions.perform()
                return

            if id and id == item.get_attribute('id'):
                self.actions.click(item)
                self.actions.perform()
                return

        raise ValueError('Item not found!')

    def has_item(self, label: str | None = None, id: str | None = None):
        if label:
            return label in self.items_labels
        elif id:
            return id in self.items_ids
        raise ValueError('You must provide at least "label" or "id"!')

    def __contains__(self, key: Any) -> bool:
        return self.has_item(label=key) or self.has_item(id=key)

    def print(self):
        identifier = f'[@id="{self.id}"]' if self.id else self.xpath
        print_list(identifier, self.items)


class ButtonElement(Element):
    @property
    def is_link(self) -> bool:
        return self.tag == 'a'

    @property
    def reference(self) -> str | None:
        if self.is_link:
            return self.attribute('href')
        return None

    def double_click(self):
        self.actions.move_to_element(self.source)
        self.actions.double_click(self.source)
        self.actions.perform()


class FormElement(Element):
    _success_redirect_url: str | None = None
    _success_elements_to_find: list[str] = []
    _success_text_to_find: list[str] = []

    def _check_success_conditions(self):
        if self._success_redirect_url is not None:
            if self.webdriver.current_url != self._success_redirect_url:
                raise FormException('redirect_url', self._success_redirect_url)

        for element in self._success_elements_to_find:
            try:
                find_element(self.webdriver, element)
            except NoSuchElementException:
                raise FormException('elements_to_find', element)

        for text in self._success_text_to_find:
            if text not in self.webdriver.page_source:
                raise FormException('text_to_find', text)

    @property
    def method(self) -> str:
        if gotten_method := self.attribute('method'):
            return gotten_method.upper()
        return 'GET'

    @property
    def action(self) -> str | None:
        return self.attribute('action')

    @property
    def type(self) -> str:
        if gotten_type := self.attribute('enctype'):
            return gotten_type
        return 'application/x-www-form-urlencoded'

    def set_success_condition(
        self,
        redirect_url: str | None = None,
        elements_to_find: list[str] = [],
        text_to_find: list[str] = [],
    ):
        self._success_redirect_url = redirect_url
        self._success_elements_to_find = elements_to_find
        self._success_text_to_find = text_to_find

    def submit(self, button_xpath: str | None = None):
        if not button_xpath:
            self.source.submit()
        else:
            button = ButtonElement(button_xpath, self.webdriver)
            button.click()
        self._check_success_conditions()


class TableElement(Element):
    @property
    def headers_sources(self) -> list[WebElement]:
        first_row = find_element(self.source, './/tr')
        return find_elements(first_row, './/th')

    @property
    def headers(self) -> list[str | None]:
        return [
            header.get_attribute('innerText') if header else None
            for header in self.headers_sources
        ]

    @property
    def rows_sources(self) -> list[WebElement]:
        rows = find_elements(self.source, './/tr')
        if self.headers:
            del rows[0]
        return rows

    @property
    def rows(self) -> list[list[str | None]]:
        rows_content = []
        for element in self.rows_sources:
            rows_content.append(
                [
                    cell.get_attribute('innerText')
                    for cell in find_elements(element, './/td | .//th')
                ]
            )
        return rows_content

    def column_values(
        self, name: str | None = None, index: int | None = None
    ) -> list[str | None]:
        if (not name) and (index is None):
            raise ValueError('You must provide at least "name" or "index"!')

        if index is None:
            index = self.headers.index(name)

        return [row[index] for row in self.rows]

    def has_content(self, value: str) -> bool:
        cells_content = [
            cell.get_attribute('innerText')
            for cell in self.source.find_elements('.//td')
        ]
        return value in cells_content

    def __contains__(self, value: Any) -> bool:
        return self.has_content(value)

    def print(self):
        print_table(self.headers, self.rows)


class ImageElement(Element):
    @property
    def text(self) -> str | None:
        return self.attribute('alt')

    @property
    def reference(self) -> str | None:
        return self.attribute('src')

    def save(self, path: str | None = None):
        if not self.reference:
            raise FileNotDownloadableException(self.reference, self.xpath)
        if path is not None:
            self.file.download_file(self.reference, path)
        else:
            self.file.download_file_to_cwd(self.reference)



from typing import Any
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.action_chains import ActionChains

from fastrpa.exceptions import KeyDoesNotExistsException
from fastrpa.types import WebDriver


class Keyboard:
    def __init__(self, webdriver: WebDriver):
        self._keys_mapping: dict[str, str] | None = None
        self._keys: list[str] | None = None
        self.webdriver = webdriver
        self.actions = ActionChains(self.webdriver)

    @property
    def keys_mapping(self) -> dict[str, str]:
        if self._keys_mapping is None:
            self._keys_mapping = {
                attr: getattr(Keys, attr)
                for attr in dir(Keys)
                if not attr.startswith('_')
            }
        return self._keys_mapping

    @property
    def keys(self) -> list[str]:
        if self._keys is None:
            self._keys = list(self.keys_mapping.keys())
        return self._keys

    def key_name(self, name: str) -> str:
        return name.upper().replace(' ', '_')

    def has_key(self, name: str | None = None, code: str | None = None) -> bool:
        if name:
            return self.key_name(name) in self.keys
        elif code:
            return code in self.keys_mapping.values()
        raise ValueError('You must provide at least "name" or "code"!')

    def key_code(self, key: str) -> str:
        key_name = self.key_name(key)
        if key_name in self.keys:
            return self.keys_mapping[key_name]
        elif key in self.keys_mapping.values():
            return key
        elif len(key) == 1:
            return key
        raise KeyDoesNotExistsException(key)

    def press(self, key: str):
        self.actions.send_keys(self.key_code(key))
        self.actions.perform()

    def shortcut(self, *keys: str):
        for key in keys:
            self.actions.key_down(self.key_code(key))
        for key in keys[::-1]:
            self.actions.key_up(self.key_code(key))
        self.actions.perform()

    def __contains__(self, value: Any) -> bool:
        return self.has_key(value) or self.has_key(code=value)



from typing import Type, TypeVar, Union
from requests import Session
from selenium.webdriver import Remote, ChromeOptions
from selenium.webdriver.common.by import By
from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
)
from fastrpa.utils import (
    find_element,
    get_browser_options,
    get_domain,
    get_session,
)
from fastrpa.core.console import Console
from fastrpa.core.cookies import Cookies
from fastrpa.core.screenshot import Screenshot
from fastrpa.core.tabs import Tabs
from fastrpa.exceptions import ElementNotCompatibleException
from fastrpa.core.elements import (
    CheckboxElement,
    Element,
    ImageElement,
    InputElement,
    FileInputElement,
    ButtonElement,
    FormElement,
    ListElement,
    RadioInputElement,
    TableElement,
    SelectElement,
)

from fastrpa.core.wait import Wait
from fastrpa.core.keyboard import Keyboard
from fastrpa.core.elements_factory import ElementsFactory
from fastrpa.types import BrowserOptions, BrowserOptionsClass, WebDriver


SpecificElement = TypeVar('SpecificElement', bound=Element)
GenericElement = Union[Element, SpecificElement]


class Web:
    def __init__(
        self,
        webdriver: WebDriver,
        timeout: int,
    ):
        self.webdriver = webdriver
        self.timeout = timeout
        self.keyboard = Keyboard(self.webdriver)
        self.screenshot = Screenshot(self.webdriver)
        self.cookies = Cookies(self.webdriver)
        self.console = Console(self.webdriver)
        self.tabs = Tabs(self.webdriver)
        self.wait = Wait(self.webdriver, self.timeout)
        self.factory = ElementsFactory(self.webdriver, self.timeout)

    @property
    def url(self) -> str:
        return self.webdriver.current_url

    @property
    def domain(self) -> str:
        return get_domain(self.webdriver)

    @property
    def title(self) -> str:
        return self.webdriver.title

    @property
    def html(self) -> str:
        return self.webdriver.page_source

    @property
    def session(self) -> Session:
        return get_session(self.webdriver)

    def browse(self, url: str):
        self.webdriver.get(url)

    def refresh(self):
        self.webdriver.refresh()

    def is_interactive(self, xpath: str) -> bool:
        try:
            if element := self.webdriver.find_element(By.XPATH, xpath):
                return element.is_displayed()
            return False
        except NoSuchElementException:
            return False
        except StaleElementReferenceException:
            return False

    def read(self, xpath: str) -> str | None:
        try:
            if element := find_element(self.webdriver, xpath):
                return element.text if element.text else None
            return None
        except NoSuchElementException:
            return None

    def element(self, xpath: str, wait: bool = True) -> GenericElement:
        if not wait:
            return self.factory.get(xpath)
        return self.factory.get_when_available(xpath)

    def elements(self, xpath: str) -> list[GenericElement]:
        return self.factory.get_many(xpath)

    def _specific_element(
        self, xpath: str, class_name: Type[SpecificElement], wait: bool = True
    ) -> SpecificElement:
        element = self.element(xpath, wait)
        if not isinstance(element, class_name):
            raise ElementNotCompatibleException(xpath, class_name)
        return element

    def input(self, xpath: str, wait: bool = True) -> InputElement:
        return self._specific_element(xpath, InputElement, wait)

    def file_input(self, xpath: str, wait: bool = True) -> FileInputElement:
        return self._specific_element(xpath, FileInputElement, wait)

    def radio_input(self, xpath: str, wait: bool = True) -> RadioInputElement:
        return self._specific_element(xpath, RadioInputElement, wait)

    def checkbox(self, xpath: str, wait: bool = True) -> CheckboxElement:
        return self._specific_element(xpath, CheckboxElement, wait)

    def select(self, xpath: str, wait: bool = True) -> SelectElement:
        return self._specific_element(xpath, SelectElement, wait)

    def button(self, xpath: str, wait: bool = True) -> ButtonElement:
        return self._specific_element(xpath, ButtonElement, wait)

    def form(self, xpath: str, wait: bool = True) -> FormElement:
        return self._specific_element(xpath, FormElement, wait)

    def list(self, xpath: str, wait: bool = True) -> ListElement:
        return self._specific_element(xpath, ListElement, wait)

    def table(self, xpath: str, wait: bool = True) -> TableElement:
        return self._specific_element(xpath, TableElement, wait)

    def image(self, xpath: str, wait: bool = True) -> ImageElement:
        return self._specific_element(xpath, ImageElement, wait)


class FastRPA:
    browser_arguments = ['--start-maximized', '--ignore-certificate-errors']

    def __init__(
        self,
        webdriver: WebDriver | None = None,
        options_class: BrowserOptionsClass = ChromeOptions,
        browser_arguments: list[str] | None = None,
        timeout: int = 15,
    ):
        self._browser_options: BrowserOptions | None = None
        self._webdriver = webdriver
        self._options_class = options_class
        self.timeout = timeout

        if browser_arguments:
            self.browser_arguments = browser_arguments

    @property
    def browser_options(self) -> BrowserOptions:
        if self._browser_options is None:
            self._browser_options = get_browser_options(
                options=self.browser_arguments,
                options_class=self._options_class,
            )
        return self._browser_options

    @property
    def webdriver(self) -> WebDriver:
        if self._webdriver is None:
            self._webdriver = Remote(options=self.browser_options)
        return self._webdriver

    def __del__(self):
        self.webdriver.quit()

    def web(self, url: str | None = None) -> Web:
        obj = Web(self.webdriver, self.timeout)
        if url is not None:
            obj.browse(url)
        return obj



from dataclasses import dataclass
from typing import Any, Literal, Type
from selenium.webdriver import (
    ChromeOptions,
    SafariOptions,
    FirefoxOptions,
    Chrome,
    Safari,
    Firefox,
    Remote,
)


WebDriver = Remote | Chrome | Safari | Firefox
BrowserOptions = ChromeOptions | SafariOptions | FirefoxOptions
BrowserOptionsClass = (
    Type[ChromeOptions] | Type[SafariOptions] | Type[FirefoxOptions]
)


@dataclass
class Cookie:
    name: str
    value: str
    domain: str
    path: str = '/'
    secure: bool = False
    http_only: bool = True
    same_site: Literal['Strict', 'Lax', 'None'] = 'None'

    @staticmethod
    def from_selenium(content: dict[str, Any]) -> 'Cookie':
        return Cookie(
            name=content['name'],
            value=content['value'],
            domain=content['domain'],
            path=content['path'],
            secure=content['secure'],
            http_only=content['httpOnly'],
            same_site=content['sameSite'],
        )

    def to_selenium(self) -> dict[str, Any]:
        return {
            'name': self.name,
            'value': self.value,
            'domain': self.domain,
            'path': self.path,
            'secure': self.secure,
            'httpOnly': self.http_only,
            'sameSite': self.same_site,
        }

    def to_requests(self) -> dict[str, Any]:
        return {
            'name': self.name,
            'value': self.value,
            'domain': self.domain,
            'path': self.path,
            'secure': self.secure,
        }



from datetime import datetime
from selenium.webdriver.common.by import By
from fastrpa.exceptions import ScreenshotNotTakenException
from fastrpa.types import WebDriver


class Screenshot:
    def __init__(self, webdriver: WebDriver):
        self.webdriver = webdriver

    def _file_path(self, path: str | None = None) -> str:
        if path is None:
            path = f'{datetime.now().isoformat()}.png'
        return path

    def _restore_window(
        self, size: dict[str, int], position: dict[str, int], is_maximized: bool
    ):
        if not is_maximized:
            self.webdriver.set_window_size(size['width'], size['height'])
            self.webdriver.set_window_position(position['x'], position['y'])
        else:
            self.webdriver.maximize_window()

    @property
    def max_width(self) -> int:
        return self.webdriver.execute_script(
            'return document.documentElement.scrollWidth'
        )

    @property
    def max_height(self) -> int:
        return self.webdriver.execute_script(
            'return document.documentElement.scrollHeight + (window.innerHeight * 0.2)'
        )

    @property
    def is_maximized(self) -> bool:
        window_size = self.webdriver.get_window_size()
        screen_width = self.webdriver.execute_script('return screen.availWidth')
        screen_height = self.webdriver.execute_script(
            'return screen.availHeight'
        )
        return (window_size['width'] == screen_width) and (
            window_size['height'] == screen_height
        )

    @property
    def image(self) -> bytes:
        return self.webdriver.get_screenshot_as_png()

    @property
    def full_page_image(self) -> bytes:
        starter_position = self.webdriver.get_window_position()
        starter_size = self.webdriver.get_window_size()
        starter_is_maximized = self.is_maximized
        image: bytes

        try:
            self.webdriver.set_window_size(self.max_width, self.max_height)
            element = self.webdriver.find_element(By.XPATH, '//body')
            image = element.screenshot_as_png

        finally:
            self._restore_window(
                size=starter_size,
                position=starter_position,
                is_maximized=starter_is_maximized,
            )
            return image

    def save_image(self, path: str | None = None):
        success = self.webdriver.save_screenshot(self._file_path(path))
        if not success:
            raise ScreenshotNotTakenException(
                'image', self.webdriver.current_url
            )

    def save_full_page_image(self, path: str | None = None):
        starter_size = self.webdriver.get_window_size()
        starter_position = self.webdriver.get_window_position()
        starter_is_maximized = self.is_maximized

        try:
            self.webdriver.set_window_size(self.max_width, self.max_height)
            element = self.webdriver.find_element(By.XPATH, '//body')
            success = element.screenshot(self._file_path(path))

            if not success:
                raise ScreenshotNotTakenException(
                    'image', self.webdriver.current_url
                )

        finally:
            self._restore_window(
                size=starter_size,
                position=starter_position,
                is_maximized=starter_is_maximized,
            )



from time import sleep

from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from fastrpa.types import WebDriver


class Wait:
    def __init__(self, webdriver: WebDriver, timeout: float):
        self._timeout = timeout
        self.webdriver = webdriver

    def _get_timeout(self, value: float | None) -> float:
        if value is not None:
            return value
        return self._timeout

    def source(self, timeout: float | None = None) -> WebDriverWait:
        return WebDriverWait(self.webdriver, self._get_timeout(timeout))

    def seconds(self, seconds: float):
        sleep(seconds)

    def is_present(self, xpath: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until(
            EC.presence_of_element_located((By.XPATH, xpath))
        )

    def not_is_present(self, xpath: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until_not(
            EC.presence_of_element_located((By.XPATH, xpath))
        )

    def is_clickable(self, xpath: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until(
            EC.element_to_be_clickable((By.XPATH, xpath))
        )

    def not_is_clickable(self, xpath: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until_not(
            EC.element_to_be_clickable((By.XPATH, xpath))
        )

    def is_hidden(self, xpath: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until(
            EC.invisibility_of_element_located((By.XPATH, xpath))
        )

    def not_is_hidden(self, xpath: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until_not(
            EC.invisibility_of_element_located((By.XPATH, xpath))
        )

    def contains_text(
        self, xpath: str, text: str, timeout: float | None = None
    ):
        self.source(self._get_timeout(timeout)).until(
            EC.text_to_be_present_in_element((By.XPATH, xpath), text)
        )

    def not_contains_text(
        self, xpath: str, text: str, timeout: float | None = None
    ):
        self.source(self._get_timeout(timeout)).until_not(
            EC.text_to_be_present_in_element((By.XPATH, xpath), text)
        )

    def url_contains(self, sub_url: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until(EC.url_contains(sub_url))

    def not_url_contains(self, sub_url: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until_not(
            EC.url_contains(sub_url)
        )

    def title_contains(self, sub_title: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until(
            EC.title_contains(sub_title)
        )

    def not_title_contains(self, sub_title: str, timeout: float | None = None):
        self.source(self._get_timeout(timeout)).until_not(
            EC.title_contains(sub_title)
        )



class ElementNotFoundException(Exception):
    message = 'No one element [{}] was found!'

    def __init__(self, xpath: str):
        self.xpath = xpath
        super().__init__(self.message.format(xpath))


class ElementTimeoutException(Exception):
    message = 'Element [{}] not found after {} seconds!'

    def __init__(self, xpath: str, timeout: float):
        self.xpath = xpath
        self.timeout = timeout
        super().__init__(self.message.format(xpath, timeout))


class ElementNotCompatibleException(Exception):
    message = 'Element [{}] is not compatible with {}!'

    def __init__(self, xpath: str, class_name: type):
        self.xpath = xpath
        self.class_name = class_name
        super().__init__(self.message.format(xpath, class_name))


class ScreenshotNotTakenException(Exception):
    message = 'The [{}] screenshot from [{}] was not taken!'

    def __init__(self, type: str, url: str):
        self.type = type
        self.url = url
        super().__init__(self.message.format(type, url))


class CookieNotAddedException(Exception):
    message = 'The cookie [{}] was not added!'

    def __init__(self, cookie_name: str):
        self.cookie_name = cookie_name
        super().__init__(self.message.format(cookie_name))


class KeyDoesNotExistsException(Exception):
    message = 'The key [{}] does not exists!'

    def __init__(self, key: str):
        self.key = key
        super().__init__(self.message.format(key))


class FormException(Exception):
    message = (
        'The form submission got an error! Condition [{}, {}] not satisfied!'
    )

    def __init__(self, condition: str, expected_value: str):
        self.condition = condition
        self.expected_value = expected_value
        super().__init__(self.message.format(condition, expected_value))


class FileNotDownloadableException(Exception):
    message = 'The file [{}] at [{}] is not downloadable!'

    def __init__(self, url: str | None, xpath: str):
        self.url = url
        self.xpath = xpath
        super().__init__(self.message.format(url, xpath))



from typing import Type
from selenium.common.exceptions import TimeoutException, NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.remote.webelement import WebElement

from fastrpa.core.elements import (
    ButtonElement,
    CheckboxElement,
    Element,
    FileInputElement,
    FormElement,
    ImageElement,
    InputElement,
    ListElement,
    RadioInputElement,
    SelectElement,
    TableElement,
)
from fastrpa.utils import ensure_element
from fastrpa.exceptions import ElementNotFoundException, ElementTimeoutException
from fastrpa.types import WebDriver


class ElementsFactory:
    def __init__(self, webdriver: WebDriver, timeout: int):
        self.timeout = timeout
        self.webdriver = webdriver

    def element_class(self, element: WebElement) -> Type[Element]:
        if element.tag_name in ['a', 'button']:
            return ButtonElement

        if element.tag_name in ['ol', 'ul']:
            return ListElement

        elif all(
            [
                element.tag_name == 'input',
                element.get_attribute('type') == 'file',
            ]
        ):
            return FileInputElement

        elif all(
            [
                element.tag_name == 'input',
                element.get_attribute('type') == 'radio',
            ]
        ):
            return RadioInputElement

        elif all(
            [
                element.tag_name == 'input',
                element.get_attribute('type') == 'checkbox',
            ]
        ):
            return CheckboxElement

        elif element.tag_name in ['input', 'textarea']:
            return InputElement

        elif element.tag_name == 'select':
            return SelectElement

        elif element.tag_name == 'form':
            return FormElement

        elif element.tag_name == 'table':
            return TableElement

        elif element.tag_name == 'img':
            return ImageElement

        return Element

    @ensure_element
    def get_when_available(self, xpath: str) -> Element:
        try:
            selenium_element = WebDriverWait(
                self.webdriver, self.timeout
            ).until(
                expected_conditions.presence_of_element_located(
                    (By.XPATH, xpath)
                )
            )
            element_class = self.element_class(selenium_element)
            return element_class(xpath, self.webdriver)

        except TimeoutException:
            raise ElementTimeoutException(xpath, self.timeout)

    @ensure_element
    def get(self, xpath: str) -> Element:
        try:
            selenium_element = self.webdriver.find_element(By.XPATH, xpath)
            element_class = self.element_class(selenium_element)
            return element_class(xpath, self.webdriver)

        except NoSuchElementException:
            raise ElementNotFoundException(xpath)

    @ensure_element
    def get_many(self, xpath: str) -> list[Element]:
        elements_to_return = []

        try:
            for selenium_element in self.webdriver.find_elements(
                By.XPATH, xpath
            ):
                element_class = self.element_class(selenium_element)
                elements_to_return.append(element_class(xpath, self.webdriver))

            return elements_to_return

        except NoSuchElementException:
            raise ElementNotFoundException(xpath)



from fastrpa.app import FastRPA, Web
from fastrpa.core.elements import (
    Element,
    InputElement,
    FileInputElement,
    RadioInputElement,
    CheckboxElement,
    SelectElement,
    ListElement,
    ButtonElement,
    FormElement,
    TableElement,
)


__all__ = (
    'FastRPA',
    'Web',
    'Element',
    'InputElement',
    'FileInputElement',
    'RadioInputElement',
    'CheckboxElement',
    'SelectElement',
    'ListElement',
    'ButtonElement',
    'FormElement',
    'TableElement',
)



from typing import Any

from fastrpa.utils import get_domain
from fastrpa.exceptions import CookieNotAddedException
from fastrpa.types import WebDriver, Cookie


class Cookies:
    def __init__(self, webdriver: WebDriver):
        self.webdriver = webdriver

    @property
    def list_names(self) -> list[str]:
        cookies = []
        for cookie in self.webdriver.get_cookies():
            cookies.append(cookie['name'])
        return cookies

    @property
    def list(self) -> list[Cookie]:
        cookies = []
        for cookie in self.webdriver.get_cookies():
            cookies.append(Cookie.from_selenium(cookie))
        return cookies

    def get(self, name: str) -> Cookie | None:
        if cookie := self.webdriver.get_cookie(name):
            return Cookie.from_selenium(cookie)
        return None

    def add(self, name: str, value: str, secure: bool = False) -> Cookie:
        domain = get_domain(self.webdriver)
        cookie = Cookie(name, value, domain, secure=secure)
        self.webdriver.add_cookie(cookie.to_selenium())
        if added_cookie := self.webdriver.get_cookie(name):
            return Cookie.from_selenium(added_cookie)
        raise CookieNotAddedException(name)

    def delete(self, name: str):
        self.webdriver.delete_cookie(name)

    def check(self, name: str, value: str) -> bool:
        if cookie := self.get(name):
            return cookie.value == value
        return False

    def has_cookie(self, name: str) -> bool:
        return name in self.list_names

    def __contains__(self, key: Any) -> bool:
        return key in self.list_names



def attribute_contains(
    element: str, attribute: str, value: str, child: str = ''
) -> str:
    return f'//{element}[contains({attribute}, "{value}")]{child}'


def attribute_equals(
    element: str, attribute: str, value: str, child: str = ''
) -> str:
    return f'//{element}[{attribute}="{value}"]{child}'


def id_contains(value: str, child: str = '') -> str:
    return attribute_contains('*', '@id', value, child)


def id_equals(value: str, child: str = '') -> str:
    return attribute_equals('*', '@id', value, child)


def class_contains(value: str, child: str = '') -> str:
    return attribute_contains('*', '@class', value, child)


def class_equals(value: str, child: str = '') -> str:
    return attribute_equals('*', '@class', value, child)


def name_contains(value: str, child: str = '') -> str:
    return attribute_contains('*', '@name', value, child)


def name_equals(value: str, child: str = '') -> str:
    return attribute_equals('*', '@name', value, child)


def text_contains(value: str, child: str = '') -> str:
    return attribute_contains('*', 'text()', value, child)


def text_equals(value: str, child: str = '') -> str:
    return attribute_equals('*', 'text()', value, child)



import os

from mimetypes import guess_extension
from requests import Response
from tempfile import mkstemp

from fastrpa.types import WebDriver
from fastrpa.utils import get_session


class File:
    def __init__(self, webdriver: WebDriver) -> None:
        self.webdriver = webdriver

    def get_extension(self, response: Response) -> str:
        content_type = response.headers['Content-Type']
        mime_type, *_ = content_type.split(';')
        if extension := guess_extension(mime_type):
            return extension
        raise ValueError('The response does not have a Content-Type header!')

    def get_hash(self, response: Response) -> str:
        return str(abs(hash(response.content)))

    def download_file(self, url: str, path: str | None = None) -> str:
        response = get_session(self.webdriver).get(url)
        file_suffix = self.get_hash(response) + self.get_extension(response)

        if path is None:
            _, path = mkstemp(file_suffix)

        elif '.' not in path:
            path = os.path.join(path, file_suffix)

        with open(path, 'wb') as file:
            file.write(response.content)

        return path

    def download_file_to_cwd(self, url: str) -> str:
        return self.download_file(url, os.getcwd())

    def get_path(self, path: str) -> str:
        if os.path.isfile(path):
            return path
        return self.download_file(path)



from typing import Any
from fastrpa.types import WebDriver


class Console:
    def __init__(self, webdriver: WebDriver) -> None:
        self.webdriver = webdriver

    def evaluate(self, code: str) -> Any:
        return self.webdriver.execute_script(f'return {code}')

    def run(self, lines: list[str]) -> Any:
        return self.webdriver.execute_script('\n'.join(lines))

    def run_script(self, path: str) -> Any:
        with open(path, 'r') as script:
            return self.run(script.readlines())



from typing import Any
from fastrpa.types import WebDriver


class Tabs:
    def __init__(self, webdriver: WebDriver):
        self.webdriver = webdriver

    @property
    def list(self) -> list[str]:
        return self.webdriver.window_handles

    @property
    def current(self) -> str:
        return self.webdriver.current_window_handle

    @property
    def current_index(self) -> int:
        return self.list.index(self.current)

    @property
    def count(self) -> int:
        return len(self.list)

    def new(self) -> str:
        self.webdriver.switch_to.new_window('tab')
        return self.list[-1]

    def switch(self, tab: str | None = None, index: int | None = None):
        if tab:
            self.webdriver.switch_to.window(tab)
        elif index is not None:
            self.webdriver.switch_to.window(self.list[index])
        else:
            raise ValueError('You must provide at least "tab" or "index"!')

    def close(self):
        current_tab_index = self.current_index
        if self.count >= 2:
            self.webdriver.close()
            self.switch(index=current_tab_index - 1)
        else:
            self.new()
            self.switch(index=current_tab_index)
            self.webdriver.close()
            self.switch(index=current_tab_index)

    def __len__(self) -> int:
        return self.count

    def __contains__(self, value: Any) -> bool:
        return value in self.list



from typing import Callable, Iterable
from urllib.parse import urlparse
from selenium.webdriver import ChromeOptions
from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.by import By
from requests import Session

from fastrpa.types import BrowserOptions, BrowserOptionsClass, WebDriver


def get_browser_options(
    options: list[str], options_class: BrowserOptionsClass = ChromeOptions
) -> BrowserOptions:
    instance = options_class()
    for opt in options:
        instance.add_argument(opt)
    return instance


def get_session(webdriver: WebDriver) -> Session:
    obj = Session()
    for cookie in webdriver.get_cookies():
        obj.cookies.set(
            name=cookie['name'],
            value=cookie['value'],
            domain=cookie['domain'],
            path=cookie['path'],
            secure=cookie['secure'],
        )
    return obj


def get_domain(webdriver: WebDriver) -> str:
    return urlparse(webdriver.current_url).netloc


def find_element(browsable: WebDriver | WebElement, xpath: str) -> WebElement:
    return browsable.find_element(By.XPATH, xpath)


def find_elements(
    browsable: WebDriver | WebElement, xpath: str
) -> list[WebElement]:
    return browsable.find_elements(By.XPATH, xpath)


def print_table(headers: Iterable[str], rows: Iterable[str]):
    try:
        from rich.table import Table
        from rich.console import Console

        rich_table = Table(*headers)
        for row in rows:
            rich_table.add_row(*row)
        Console().print(rich_table)

    except ImportError:
        raise EnvironmentError('You need to install rich to print tables.')


def print_list(name: str, values_dict: dict[str | None, str | None]):
    try:
        from rich.tree import Tree
        from rich.console import Console

        rich_tree = Tree(name.replace('[', '\['))
        for id, value in values_dict.items():
            rich_tree.add(f'\[{id}] {value}')
        Console().print(rich_tree)

    except ImportError:
        raise EnvironmentError('You need to install rich to print lists.')


def ensure_element(func: Callable, max_attempts: int = 3):
    def wrapper(*args, **kwargs):
        attempt = 0
        while attempt < max_attempts:
            try:
                return func(*args, **kwargs)
            except StaleElementReferenceException:
                attempt += 1

    return wrapper
